import inspect
import os
import threading
import time
from concurrent.futures import Future, TimeoutError
from functools import wraps
from unittest.mock import Mock

import pytest
from qtpy.QtCore import QCoreApplication, QObject, QThread, Signal

from superqt.utils import ensure_main_thread, ensure_object_thread

skip_on_ci = pytest.mark.skipif(bool(os.getenv("CI")), reason="github hangs")


class SampleObject(QObject):
    assigment_done = Signal()

    def __init__(self):
        super().__init__()
        self.main_thread_res = {}
        self.object_thread_res = {}
        self.main_thread_prop_val = None
        self.sample_thread_prop_val = None

    def long_wait(self):
        time.sleep(1)

    @property
    def sample_main_thread_property(self):
        return self.main_thread_prop_val

    @sample_main_thread_property.setter  # type: ignore
    @ensure_main_thread()
    def sample_main_thread_property(self, value):
        if QThread.currentThread() is not QCoreApplication.instance().thread():
            raise RuntimeError("Wrong thread")
        self.main_thread_prop_val = value
        self.assigment_done.emit()

    @property
    def sample_object_thread_property(self):
        return self.sample_thread_prop_val

    @sample_object_thread_property.setter  # type: ignore
    @ensure_object_thread()
    def sample_object_thread_property(self, value):
        if QThread.currentThread() is not self.thread():
            raise RuntimeError("Wrong thread")
        self.sample_thread_prop_val = value
        self.assigment_done.emit()

    @ensure_main_thread
    def check_main_thread(self, a, *, b=1):
        if QThread.currentThread() is not QCoreApplication.instance().thread():
            raise RuntimeError("Wrong thread")
        self.main_thread_res = {"a": a, "b": b}
        self.assigment_done.emit()

    @ensure_object_thread
    def check_object_thread(self, a, *, b=1):
        if QThread.currentThread() is not self.thread():
            raise RuntimeError("Wrong thread")
        self.object_thread_res = {"a": a, "b": b}
        self.assigment_done.emit()

    @ensure_object_thread(await_return=True)
    def check_object_thread_return(self, a):
        if QThread.currentThread() is not self.thread():
            raise RuntimeError("Wrong thread")
        return a * 7

    @ensure_object_thread(await_return=True, timeout=200)
    def check_object_thread_return_timeout(self, a):
        if QThread.currentThread() is not self.thread():
            raise RuntimeError("Wrong thread")
        time.sleep(1)
        return a * 7

    @ensure_object_thread(await_return=False)
    def check_object_thread_return_future(self, a: int):
        """sample docstring"""
        if QThread.currentThread() is not self.thread():
            raise RuntimeError("Wrong thread")
        time.sleep(0.4)
        return a * 7

    @ensure_main_thread(await_return=True)
    def check_main_thread_return(self, a):
        if QThread.currentThread() is not QCoreApplication.instance().thread():
            raise RuntimeError("Wrong thread")
        return a * 8


class LocalThread(QThread):
    def __init__(self, ob):
        super().__init__()
        self.ob = ob

    def run(self):
        assert QThread.currentThread() is not QCoreApplication.instance().thread()
        self.ob.check_main_thread(5, b=8)
        self.ob.main_thread_prop_val = "text2"


class LocalThread2(QThread):
    def __init__(self, ob):
        super().__init__()
        self.ob = ob
        self.executed = False

    def run(self):
        assert QThread.currentThread() is not QCoreApplication.instance().thread()
        assert self.ob.check_main_thread_return(5) == 40
        self.executed = True


def test_only_main_thread(qapp):
    ob = SampleObject()
    ob.check_main_thread(1, b=3)
    assert ob.main_thread_res == {"a": 1, "b": 3}
    ob.check_object_thread(2, b=4)
    assert ob.object_thread_res == {"a": 2, "b": 4}
    ob.sample_main_thread_property = 5
    assert ob.sample_main_thread_property == 5
    ob.sample_object_thread_property = 7
    assert ob.sample_object_thread_property == 7


def test_main_thread(qtbot):
    ob = SampleObject()
    t = LocalThread(ob)
    with qtbot.waitSignal(t.finished):
        t.start()

    assert ob.main_thread_res == {"a": 5, "b": 8}
    assert ob.sample_main_thread_property == "text2"


def test_main_thread_return(qtbot):
    ob = SampleObject()
    t = LocalThread2(ob)
    with qtbot.wait_signal(t.finished):
        t.start()
    assert t.executed


def test_names(qapp):
    ob = SampleObject()
    assert ob.check_object_thread.__name__ == "check_object_thread"
    assert ob.check_object_thread_return.__name__ == "check_object_thread_return"
    assert (
        ob.check_object_thread_return_timeout.__name__
        == "check_object_thread_return_timeout"
    )
    assert (
        ob.check_object_thread_return_future.__name__
        == "check_object_thread_return_future"
    )
    assert ob.check_object_thread_return_future.__doc__ == "sample docstring"
    signature = inspect.signature(ob.check_object_thread_return_future)
    assert len(signature.parameters) == 1
    assert next(iter(signature.parameters.values())).name == "a"
    assert next(iter(signature.parameters.values())).annotation == int
    assert ob.check_main_thread_return.__name__ == "check_main_thread_return"


@skip_on_ci
def test_object_thread_return(qtbot):
    ob = SampleObject()
    thread = QThread()
    thread.start()
    ob.moveToThread(thread)
    assert ob.check_object_thread_return(2) == 14
    assert ob.thread() is thread
    with qtbot.waitSignal(thread.finished):
        thread.quit()


@skip_on_ci
def test_object_thread_return_timeout(qtbot):
    ob = SampleObject()
    thread = QThread()
    thread.start()
    ob.moveToThread(thread)
    with pytest.raises(TimeoutError):
        ob.check_object_thread_return_timeout(2)
    with qtbot.waitSignal(thread.finished):
        thread.quit()


@skip_on_ci
def test_object_thread_return_future(qtbot):
    ob = SampleObject()
    thread = QThread()
    thread.start()
    ob.moveToThread(thread)
    future = ob.check_object_thread_return_future(2)
    assert isinstance(future, Future)
    assert future.result() == 14
    with qtbot.waitSignal(thread.finished):
        thread.quit()


@skip_on_ci
def test_object_thread(qtbot):
    ob = SampleObject()
    thread = QThread()
    thread.start()
    ob.moveToThread(thread)
    with qtbot.waitSignal(ob.assigment_done):
        ob.check_object_thread(2, b=4)
    assert ob.object_thread_res == {"a": 2, "b": 4}

    with qtbot.waitSignal(ob.assigment_done):
        ob.sample_object_thread_property = "text"

    assert ob.sample_object_thread_property == "text"
    assert ob.thread() is thread
    with qtbot.waitSignal(thread.finished):
        thread.quit()


@pytest.mark.parametrize("mode", ["method", "func", "wrapped"])
@pytest.mark.parametrize("deco", [ensure_main_thread, ensure_object_thread])
def test_ensure_thread_sig_inspection(deco, mode):
    class Emitter(QObject):
        sig = Signal(int, int, int)

    obj = Emitter()
    mock = Mock()

    if mode == "method":

        class Receiver(QObject):
            @deco
            def func(self, a: int, b: int):
                mock(a, b)

        r = Receiver()
        obj.sig.connect(r.func)
    elif deco == ensure_object_thread:
        return  # not compatible with function types

    elif mode == "wrapped":

        def wr(fun):
            @wraps(fun)
            def wr2(*args):
                mock(*args)
                return fun(*args) * 2

            return wr2

        @deco
        @wr
        def wrapped_func(a, b):
            return a + b

        obj.sig.connect(wrapped_func)

    elif mode == "func":

        @deco
        def func(a: int, b: int) -> None:
            mock(a, b)

        obj.sig.connect(func)

    # this is the crux of the test...
    # we emit 3 args, but the function only takes 2
    # this should normally work fine in Qt.
    # testing here that the decorator doesn't break it.
    obj.sig.emit(1, 2, 3)
    mock.assert_called_once_with(1, 2)


def test_main_thread_function(qtbot):
    """Testing decorator on a function rather than QObject method."""

    mock = Mock()

    class Emitter(QObject):
        sig = Signal(int, int, int)

    @ensure_main_thread
    def func(x: int) -> None:
        mock(x, QThread.currentThread())

    e = Emitter()
    e.sig.connect(func)

    with qtbot.waitSignal(e.sig):
        thread = threading.Thread(target=e.sig.emit, args=(1, 2, 3))
        thread.start()
        thread.join()

    mock.assert_called_once_with(1, QCoreApplication.instance().thread())



from pathlib import Path

import pytest
from qtpy.QtGui import QIcon, QPixmap
from qtpy.QtWidgets import QPushButton

from superqt.fonticon import icon, pulse, setTextIcon, spin
from superqt.fonticon._qfont_icon import QFontIconStore, _ensure_identifier

TEST_PREFIX = "ico"
TEST_CHARNAME = "smiley"
TEST_CHAR = "\ue900"
TEST_GLYPHKEY = f"{TEST_PREFIX}.{TEST_CHARNAME}"
FONT_FILE = Path(__file__).parent / "fixtures" / "fake_plugin" / "icontest.ttf"


@pytest.fixture
def store(qapp):
    store = QFontIconStore().instance()
    yield store
    store.clear()


@pytest.fixture
def full_store(store):
    store.addFont(str(FONT_FILE), TEST_PREFIX, {TEST_CHARNAME: TEST_CHAR})
    return store


def test_no_font_key():
    with pytest.raises(KeyError) as err:
        icon(TEST_GLYPHKEY)
        assert "Unrecognized font key: {TEST_PREFIX!r}." in str(err)


def test_no_charmap(store):
    store.addFont(str(FONT_FILE), TEST_PREFIX)
    with pytest.raises(KeyError) as err:
        icon(TEST_GLYPHKEY)
        assert "No charmap registered for" in str(err)


def test_font_icon_works(full_store):
    icn = icon(TEST_GLYPHKEY)
    assert isinstance(icn, QIcon)
    assert isinstance(icn.pixmap(40, 40), QPixmap)

    icn = icon(f"{TEST_PREFIX}.{TEST_CHAR}")  # also works with unicode key
    assert isinstance(icn, QIcon)
    assert isinstance(icn.pixmap(40, 40), QPixmap)

    with pytest.raises(ValueError) as err:
        icon(f"{TEST_PREFIX}.smelly")  # bad name
        assert "Font 'test (Regular)' has no glyph with the key 'smelly'" in str(err)


def test_on_button(full_store, qtbot):
    btn = QPushButton(None)
    qtbot.addWidget(btn)
    btn.setIcon(icon(TEST_GLYPHKEY))


def test_btn_text_icon(full_store, qtbot):
    btn = QPushButton(None)
    qtbot.addWidget(btn)
    setTextIcon(btn, TEST_GLYPHKEY)
    assert btn.text() == TEST_CHAR


def test_animation(full_store, qtbot):
    btn = QPushButton(None)
    qtbot.addWidget(btn)
    icn = icon(TEST_GLYPHKEY, animation=pulse(btn))
    btn.setIcon(icn)
    with qtbot.waitSignal(icn._engine._default_opts.animation.timer.timeout):
        icn.pixmap(40, 40)
        btn.update()


def test_multistate(full_store, qtbot, qapp):
    """complicated multistate icon"""
    btn = QPushButton()
    qtbot.addWidget(btn)
    icn = icon(
        TEST_GLYPHKEY,
        color="blue",
        states={
            "active": {
                "color": "red",
                "scale_factor": 0.5,
                "animation": pulse(btn),
            },
            "disabled": {
                "color": "green",
                "scale_factor": 0.8,
                "animation": spin(btn),
            },
        },
    )
    btn.setIcon(icn)
    btn.show()

    btn.setEnabled(False)
    active = icn._engine._opts[QIcon.State.Off][QIcon.Mode.Active].animation.timer
    disabled = icn._engine._opts[QIcon.State.Off][QIcon.Mode.Disabled].animation.timer

    with qtbot.waitSignal(active.timeout, timeout=1000):
        btn.setEnabled(True)
        # hack to get the signal emitted
        icn.pixmap(100, 100, QIcon.Mode.Active, QIcon.State.Off)

    assert active.isActive()
    assert not disabled.isActive()
    with qtbot.waitSignal(disabled.timeout):
        btn.setEnabled(False)
    assert disabled.isActive()

    # smoke test, paint all the states
    icn.pixmap(100, 100, QIcon.Mode.Active, QIcon.State.Off)
    icn.pixmap(100, 100, QIcon.Mode.Disabled, QIcon.State.Off)
    icn.pixmap(100, 100, QIcon.Mode.Selected, QIcon.State.Off)
    icn.pixmap(100, 100, QIcon.Mode.Normal, QIcon.State.Off)
    icn.pixmap(100, 100, QIcon.Mode.Active, QIcon.State.On)
    icn.pixmap(100, 100, QIcon.Mode.Disabled, QIcon.State.On)
    icn.pixmap(100, 100, QIcon.Mode.Selected, QIcon.State.On)
    icn.pixmap(100, 100, QIcon.Mode.Normal, QIcon.State.On)


def test_ensure_identifier():
    assert _ensure_identifier("") == ""
    assert _ensure_identifier("1a") == "_1a"
    assert _ensure_identifier("from") == "from_"
    assert _ensure_identifier("hello-world") == "hello_world"
    assert _ensure_identifier("hello_world") == "hello_world"
    assert _ensure_identifier("hello world") == "hello_world"



import os
import sys
from unittest.mock import Mock

import pytest
import qtpy
from qtpy.QtCore import QObject, QTimer, Signal
from qtpy.QtWidgets import QApplication, QErrorMessage, QMessageBox

from superqt.utils import exceptions_as_dialog, signals_blocked
from superqt.utils._util import get_max_args


def test_signal_blocker(qtbot):
    """make sure context manager signal blocker works"""

    class Emitter(QObject):
        sig = Signal()

    obj = Emitter()
    receiver = Mock()
    obj.sig.connect(receiver)

    # make sure signal works
    with qtbot.waitSignal(obj.sig):
        obj.sig.emit()

    receiver.assert_called_once()
    receiver.reset_mock()

    with signals_blocked(obj):
        obj.sig.emit()
        qtbot.wait(10)

    receiver.assert_not_called()


def test_get_max_args_simple():
    def fun1():
        pass

    assert get_max_args(fun1) == 0

    def fun2(a):
        pass

    assert get_max_args(fun2) == 1

    def fun3(a, b=1):
        pass

    assert get_max_args(fun3) == 2

    def fun4(a, *, b=2):
        pass

    assert get_max_args(fun4) == 1

    def fun5(a, *b):
        pass

    assert get_max_args(fun5) is None

    assert get_max_args(print) is None


def test_get_max_args_wrapped():
    from functools import partial, wraps

    def fun1(a, b):
        pass

    assert get_max_args(partial(fun1, 1)) == 1

    def dec(fun):
        @wraps(fun)
        def wrapper(*args, **kwargs):
            return fun(*args, **kwargs)

        return wrapper

    assert get_max_args(dec(fun1)) == 2


def test_get_max_args_methods():
    class A:
        def fun1(self):
            pass

        def fun2(self, a):
            pass

        def __call__(self, a, b=1):
            pass

    assert get_max_args(A().fun1) == 0
    assert get_max_args(A().fun2) == 1
    assert get_max_args(A()) == 2


MAC_CI_PYSIDE6 = bool(
    sys.platform == "darwin" and os.getenv("CI") and qtpy.API_NAME == "PySide6"
)


@pytest.mark.skipif(MAC_CI_PYSIDE6, reason="still hangs on mac ci with pyside6")
def test_exception_context(qtbot, qapp: QApplication) -> None:
    def accept():
        for wdg in qapp.topLevelWidgets():
            if isinstance(wdg, QMessageBox):
                wdg.button(QMessageBox.StandardButton.Ok).click()

    with exceptions_as_dialog():
        QTimer.singleShot(0, accept)
        raise Exception("This will be caught and shown in a QMessageBox")

    with pytest.raises(ZeroDivisionError), exceptions_as_dialog(ValueError):
        1 / 0  # noqa

    with exceptions_as_dialog(msg_template="Error: {exc_value}"):
        QTimer.singleShot(0, accept)
        raise Exception("This message will be used as 'exc_value'")

    err = QErrorMessage()
    with exceptions_as_dialog(use_error_message=err):
        QTimer.singleShot(0, err.accept)
        raise AssertionError("Uncheck the checkbox to ignore this in the future")

    # tb formatting smoke test, and return value checking
    exc = ValueError("Bad Val")
    with exceptions_as_dialog(
        msg_template="{tb}",
        buttons=QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel,
    ) as ctx:
        qtbot.addWidget(ctx.dialog)
        QTimer.singleShot(100, accept)
        raise exc

    assert isinstance(ctx.dialog, QMessageBox)
    assert ctx.dialog.result() == QMessageBox.StandardButton.Ok
    assert ctx.exception is exc



from typing import List, Tuple

import pytest
from pytestqt.qtbot import QtBot
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QTreeWidget, QTreeWidgetItem

from superqt import QSearchableTreeWidget


@pytest.fixture
def data() -> dict:
    return {
        "none": None,
        "str": "test",
        "int": 42,
        "list": [2, 3, 5],
        "dict": {
            "float": 0.5,
            "tuple": (22, 99),
            "bool": False,
        },
    }


@pytest.fixture
def widget(qtbot: QtBot, data: dict) -> QSearchableTreeWidget:
    widget = QSearchableTreeWidget.fromData(data)
    qtbot.addWidget(widget)
    return widget


def columns(item: QTreeWidgetItem) -> Tuple[str, str]:
    return item.text(0), item.text(1)


def all_items(tree: QTreeWidget) -> List[QTreeWidgetItem]:
    return tree.findItems("", Qt.MatchContains | Qt.MatchRecursive)


def shown_items(tree: QTreeWidget) -> List[QTreeWidgetItem]:
    items = all_items(tree)
    return [item for item in items if not item.isHidden()]


def test_init(qtbot: QtBot):
    widget = QSearchableTreeWidget()
    qtbot.addWidget(widget)
    assert widget.tree.topLevelItemCount() == 0


def test_from_data(qtbot: QtBot, data: dict):
    widget = QSearchableTreeWidget.fromData(data)
    qtbot.addWidget(widget)
    tree = widget.tree

    assert tree.topLevelItemCount() == 5

    none_item = tree.topLevelItem(0)
    assert columns(none_item) == ("none", "None")
    assert none_item.childCount() == 0

    str_item = tree.topLevelItem(1)
    assert columns(str_item) == ("str", "test")
    assert str_item.childCount() == 0

    int_item = tree.topLevelItem(2)
    assert columns(int_item) == ("int", "42")
    assert int_item.childCount() == 0

    list_item = tree.topLevelItem(3)
    assert columns(list_item) == ("list", "list")
    assert list_item.childCount() == 3
    assert columns(list_item.child(0)) == ("0", "2")
    assert columns(list_item.child(1)) == ("1", "3")
    assert columns(list_item.child(2)) == ("2", "5")

    dict_item = tree.topLevelItem(4)
    assert columns(dict_item) == ("dict", "dict")
    assert dict_item.childCount() == 3
    assert columns(dict_item.child(0)) == ("float", "0.5")
    tuple_item = dict_item.child(1)
    assert columns(tuple_item) == ("tuple", "tuple")
    assert tuple_item.childCount() == 2
    assert columns(tuple_item.child(0)) == ("0", "22")
    assert columns(tuple_item.child(1)) == ("1", "99")
    assert columns(dict_item.child(2)) == ("bool", "False")


def test_set_data(widget: QSearchableTreeWidget):
    tree = widget.tree
    assert tree.topLevelItemCount() != 1

    widget.setData({"test": "reset"})

    assert tree.topLevelItemCount() == 1
    assert columns(tree.topLevelItem(0)) == ("test", "reset")


def test_search_no_match(widget: QSearchableTreeWidget):
    widget.filter.setText("no match here")
    items = shown_items(widget.tree)
    assert len(items) == 0


def test_search_all_match(widget: QSearchableTreeWidget):
    widget.filter.setText("")
    tree = widget.tree
    assert all_items(tree) == shown_items(tree)


def test_search_match_one_key(widget: QSearchableTreeWidget):
    widget.filter.setText("int")
    items = shown_items(widget.tree)
    assert len(items) == 1
    assert columns(items[0]) == ("int", "42")


def test_search_match_one_value(widget: QSearchableTreeWidget):
    widget.filter.setText("test")
    items = shown_items(widget.tree)
    assert len(items) == 1
    assert columns(items[0]) == ("str", "test")


def test_search_match_many_keys(widget: QSearchableTreeWidget):
    widget.filter.setText("n")
    items = shown_items(widget.tree)
    assert len(items) == 2
    assert columns(items[0]) == ("none", "None")
    assert columns(items[1]) == ("int", "42")


def test_search_match_one_show_unmatched_descendants(widget: QSearchableTreeWidget):
    widget.filter.setText("list")
    items = shown_items(widget.tree)
    assert len(items) == 4
    assert columns(items[0]) == ("list", "list")
    assert columns(items[1]) == ("0", "2")
    assert columns(items[2]) == ("1", "3")
    assert columns(items[3]) == ("2", "5")


def test_search_match_one_show_unmatched_ancestors(widget: QSearchableTreeWidget):
    widget.filter.setText("tuple")
    items = shown_items(widget.tree)
    assert len(items) == 4
    assert columns(items[0]) == ("dict", "dict")
    assert columns(items[1]) == ("tuple", "tuple")
    assert columns(items[2]) == ("0", "22")
    assert columns(items[3]) == ("1", "99")



import contextlib
from typing import ClassVar, Dict, List, Set, Tuple

from ._iconfont import IconFontMeta, namespace2font

try:
    from importlib.metadata import EntryPoint, entry_points
except ImportError:
    from importlib_metadata import EntryPoint, entry_points  # type: ignore


class FontIconManager:
    ENTRY_POINT: ClassVar[str] = "superqt.fonticon"
    _PLUGINS: ClassVar[Dict[str, EntryPoint]] = {}
    _LOADED: ClassVar[Dict[str, IconFontMeta]] = {}
    _BLOCKED: ClassVar[Set[EntryPoint]] = set()

    def _discover_fonts(self) -> None:
        self._PLUGINS.clear()
        entries = entry_points()
        if hasattr(entries, "select"):  # python>3.10
            _entries = entries.select(group=self.ENTRY_POINT)  # type: ignore
        else:
            _entries = entries.get(self.ENTRY_POINT, [])
        for ep in _entries:
            if ep not in self._BLOCKED:
                self._PLUGINS[ep.name] = ep

    def _get_font_class(self, key: str) -> IconFontMeta:
        """Get IconFont given a key.

        Parameters
        ----------
        key : str
            font key to load.

        Returns
        -------
        IconFontMeta
            Instance of IconFontMeta

        Raises
        ------
        KeyError
            If no plugin provides this key
        ImportError
            If a plugin provides the key, but the entry point doesn't load
        TypeError
            If the entry point loads, but is not an IconFontMeta
        """
        if key not in self._LOADED:
            # get the entrypoint
            if key not in self._PLUGINS:
                self._discover_fonts()
            ep = self._PLUGINS.get(key)
            if ep is None:
                raise KeyError(f"No plugin provides the key {key!r}")

            # load the entry point
            try:
                font = ep.load()
            except Exception as e:
                self._PLUGINS.pop(key)
                self._BLOCKED.add(ep)
                raise ImportError(f"Failed to load {ep.value}. Plugin blocked") from e

            # make sure it's a proper IconFont
            try:
                self._LOADED[key] = namespace2font(font, ep.name.upper())
            except Exception as e:
                self._PLUGINS.pop(key)
                self._BLOCKED.add(ep)
                raise TypeError(
                    f"Failed to create fonticon from {ep.value}: {e}"
                ) from e
        return self._LOADED[key]

    def dict(self) -> dict:
        return {
            key: sorted(filter(lambda x: not x.startswith("_"), cls.__dict__))
            for key, cls in self._LOADED.items()
        }


_manager = FontIconManager()
get_font_class = _manager._get_font_class


def discover() -> Tuple[str]:
    _manager._discover_fonts()


def available() -> Tuple[str]:
    return tuple(_manager._PLUGINS)


def loaded(load_all=False) -> Dict[str, List[str]]:
    if load_all:
        discover()
        for x in available():
            with contextlib.suppress(Exception):
                _manager._get_font_class(x)
    return {
        key: sorted(filter(lambda x: not x.startswith("_"), cls.__dict__))
        for key, cls in _manager._LOADED.items()
    }



from __future__ import annotations

import warnings
from collections import abc, defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, DefaultDict, Sequence, Tuple, Union, cast

from qtpy import QT_VERSION
from qtpy.QtCore import QObject, QPoint, QRect, QSize, Qt
from qtpy.QtGui import (
    QColor,
    QFont,
    QFontDatabase,
    QGuiApplication,
    QIcon,
    QIconEngine,
    QPainter,
    QPixmap,
    QPixmapCache,
    QTransform,
)
from qtpy.QtWidgets import QApplication, QStyleOption, QWidget
from typing_extensions import TypedDict

from superqt.utils import QMessageHandler

from ._animations import Animation


class Unset:
    def __repr__(self) -> str:
        return "UNSET"


_Unset = Unset()

# A 16 pixel-high icon yields a font size of 14, which is pixel perfect
# for font-awesome. 16 * 0.875 = 14
# The reason why the glyph size is smaller than the icon size is to
# account for font bearing.
DEFAULT_SCALING_FACTOR = 0.875
DEFAULT_OPACITY = 1
ValidColor = Union[
    QColor,
    int,
    str,
    Qt.GlobalColor,
    Tuple[int, int, int, int],
    Tuple[int, int, int],
    None,
]

StateOrMode = Union[QIcon.State, QIcon.Mode]
StateModeKey = Union[StateOrMode, str, Sequence[StateOrMode]]
_SM_MAP: dict[str, StateOrMode] = {
    "on": QIcon.State.On,
    "off": QIcon.State.Off,
    "normal": QIcon.Mode.Normal,
    "active": QIcon.Mode.Active,
    "selected": QIcon.Mode.Selected,
    "disabled": QIcon.Mode.Disabled,
}


def _norm_state_mode(key: StateModeKey) -> tuple[QIcon.State, QIcon.Mode]:
    """Return state/mode tuple given a variety of valid inputs.

    Input can be either a string, or a sequence of state or mode enums.
    Strings can be any combination of on, off, normal, active, selected, disabled,
    sep by underscore.
    """
    _sm: Sequence[StateOrMode]
    if isinstance(key, str):
        try:
            _sm = [_SM_MAP[k.lower()] for k in key.split("_")]
        except KeyError as e:
            raise ValueError(
                f"{key!r} is not a valid state key, must be a combination of {{on, "
                "off, active, disabled, selected, normal} separated by underscore"
            ) from e
    else:
        _sm = key if isinstance(key, abc.Sequence) else [key]

    state = next((i for i in _sm if isinstance(i, QIcon.State)), QIcon.State.Off)
    mode = next((i for i in _sm if isinstance(i, QIcon.Mode)), QIcon.Mode.Normal)
    return state, mode


class IconOptionDict(TypedDict, total=False):
    glyph_key: str
    scale_factor: float
    color: ValidColor
    opacity: float
    animation: Animation | None
    transform: QTransform | None


# public facing, for a nicer IDE experience than a dict
# The difference between IconOpts and _IconOptions is that all of IconOpts
# all default to `_Unset` and are intended to extend some base/default option
# IconOpts are *not* guaranteed to be fully capable of rendering an icon, whereas
# IconOptions are.
@dataclass
class IconOpts:
    """Options for rendering an icon.

    Parameters
    ----------
    glyph_key : str, optional
        The key of the glyph to use, e.g. `'fa5s.smile'`, by default `None`
    scale_factor : float, optional
        The scale factor to use, by default `None`
    color : ValidColor, optional
        The color to use, by default `None`. Colors may be specified as a string,
        `QColor`, `Qt.GlobalColor`, or a 3 or 4-tuple of integers.
    opacity : float, optional
        The opacity to use, by default `None`
    animation : Animation, optional
        The animation to use, by default `None`
    """

    glyph_key: str | Unset = _Unset
    scale_factor: float | Unset = _Unset
    color: ValidColor | Unset = _Unset
    opacity: float | Unset = _Unset
    animation: Animation | Unset | None = _Unset
    transform: QTransform | Unset | None = _Unset

    def dict(self) -> IconOptionDict:
        # not using asdict due to pickle errors on animation
        d = {k: v for k, v in vars(self).items() if v is not _Unset}
        return cast(IconOptionDict, d)


@dataclass
class _IconOptions:
    """The set of options needed to render a font in a single State/Mode."""

    glyph_key: str
    scale_factor: float = DEFAULT_SCALING_FACTOR
    color: ValidColor = None
    opacity: float = DEFAULT_OPACITY
    animation: Animation | None = None
    transform: QTransform | None = None

    def _update(self, icon_opts: IconOpts) -> _IconOptions:
        return _IconOptions(**{**vars(self), **icon_opts.dict()})

    def dict(self) -> IconOptionDict:
        # not using asdict due to pickle errors on animation
        return cast(IconOptionDict, vars(self))


class _QFontIconEngine(QIconEngine):
    _opt_hash: str = ""

    def __init__(self, options: _IconOptions):
        super().__init__()
        self._opts: defaultdict[
            QIcon.State, dict[QIcon.Mode, _IconOptions | None]
        ] = DefaultDict(dict)
        self._opts[QIcon.State.Off][QIcon.Mode.Normal] = options
        self.update_hash()

    @property
    def _default_opts(self) -> _IconOptions:
        return cast(_IconOptions, self._opts[QIcon.State.Off][QIcon.Mode.Normal])

    def _add_opts(self, state: QIcon.State, mode: QIcon.Mode, opts: IconOpts) -> None:
        self._opts[state][mode] = self._default_opts._update(opts)
        self.update_hash()

    def clone(self) -> QIconEngine:  # pragma: no cover
        ico = _QFontIconEngine(self._default_opts)
        ico._opts = self._opts.copy()
        return ico

    def _get_opts(self, state: QIcon.State, mode: QIcon.Mode) -> _IconOptions:
        opts = self._opts[state].get(mode)
        if opts:
            return opts

        opp_state = QIcon.State.Off if state == QIcon.State.On else QIcon.State.On
        if mode in (QIcon.Mode.Disabled, QIcon.Mode.Selected):
            opp_mode = (
                QIcon.Mode.Disabled
                if mode == QIcon.Mode.Selected
                else QIcon.Mode.Selected
            )
            for m, s in [
                (QIcon.Mode.Normal, state),
                (QIcon.Mode.Active, state),
                (mode, opp_state),
                (QIcon.Mode.Normal, opp_state),
                (QIcon.Mode.Active, opp_state),
                (opp_mode, state),
                (opp_mode, opp_state),
            ]:
                opts = self._opts[s].get(m)
                if opts:
                    return opts
        else:
            opp_mode = (
                QIcon.Mode.Active if mode == QIcon.Mode.Normal else QIcon.Mode.Normal
            )
            for m, s in [
                (opp_mode, state),
                (mode, opp_state),
                (opp_mode, opp_state),
                (QIcon.Mode.Disabled, state),
                (QIcon.Mode.Selected, state),
                (QIcon.Mode.Disabled, opp_state),
                (QIcon.Mode.Selected, opp_state),
            ]:
                opts = self._opts[s].get(m)
                if opts:
                    return opts
        return self._default_opts

    def paint(
        self,
        painter: QPainter,
        rect: QRect,
        mode: QIcon.Mode,
        state: QIcon.State,
    ) -> None:
        opts = self._get_opts(state, mode)

        char, family, style = QFontIconStore.key2glyph(opts.glyph_key)

        # font
        font = QFont()
        font.setFamily(family)  # set separately for Qt6
        font.setPixelSize(round(rect.height() * opts.scale_factor))
        if style:
            font.setStyleName(style)

        # color
        if isinstance(opts.color, tuple):
            color_args = opts.color
        else:
            color_args = (opts.color,) if opts.color else ()

        # animation
        if opts.animation is not None:
            opts.animation.animate(painter)

        # animation
        if opts.transform is not None:
            painter.setTransform(opts.transform, True)

        painter.save()
        painter.setPen(QColor(*color_args))
        painter.setOpacity(opts.opacity)
        painter.setFont(font)
        with QMessageHandler():  # avoid "Populating font family aliases" warning
            painter.drawText(rect, Qt.AlignmentFlag.AlignCenter, char)
        painter.restore()

    def pixmap(self, size: QSize, mode: QIcon.Mode, state: QIcon.State) -> QPixmap:
        # first look in cache
        pmckey = self._pmcKey(size, mode, state)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "QPixmapCache.find")
            pm = QPixmapCache.find(pmckey) if pmckey else None
        if pm:
            return pm
        pixmap = QPixmap(size)
        if not size.isValid():
            return pixmap
        pixmap.fill(Qt.GlobalColor.transparent)
        painter = QPainter(pixmap)
        self.paint(painter, QRect(QPoint(0, 0), size), mode, state)
        painter.end()

        # Apply palette-based styles for disabled/selected modes
        # unless the user has specifically set a color for this mode/state
        if mode != QIcon.Mode.Normal:
            ico_opts = self._opts[state].get(mode)
            if not ico_opts or not ico_opts.color:
                opt = QStyleOption()
                opt.palette = QGuiApplication.palette()
                generated = QApplication.style().generatedIconPixmap(mode, pixmap, opt)
                if not generated.isNull():
                    pixmap = generated

        if pmckey and not pixmap.isNull():
            QPixmapCache.insert(pmckey, pixmap)

        return pixmap

    def _pmcKey(self, size: QSize, mode: QIcon.Mode, state: QIcon.State) -> str:
        # Qt6-style enums
        if self._get_opts(state, mode).animation:
            return ""
        if hasattr(mode, "value"):
            mode = mode.value
        if hasattr(state, "value"):
            state = state.value
        k = ((((((size.width()) << 11) | size.height()) << 11) | mode) << 4) | state
        return f"$superqt_{self._opt_hash}_{hex(k)}"

    def update_hash(self) -> None:
        hsh = id(self)
        for state, d in self._opts.items():
            for mode, opts in d.items():
                if not opts:
                    continue
                hsh += hash(
                    hash(opts.glyph_key) + hash(opts.color) + hash(state) + hash(mode)
                )
        self._opt_hash = hex(hsh)


class QFontIcon(QIcon):
    def __init__(self, options: _IconOptions) -> None:
        self._engine = _QFontIconEngine(options)
        super().__init__(self._engine)

    def addState(
        self,
        state: QIcon.State = QIcon.State.Off,
        mode: QIcon.Mode = QIcon.Mode.Normal,
        glyph_key: str | Unset = _Unset,
        scale_factor: float | Unset = _Unset,
        color: ValidColor | Unset = _Unset,
        opacity: float | Unset = _Unset,
        animation: Animation | Unset | None = _Unset,
        transform: QTransform | Unset | None = _Unset,
    ) -> None:
        """Set icon options for a specific mode/state."""
        if glyph_key is not _Unset:
            QFontIconStore.key2glyph(glyph_key)  # type: ignore

        _opts = IconOpts(
            glyph_key=glyph_key,
            scale_factor=scale_factor,
            color=color,
            opacity=opacity,
            animation=animation,
            transform=transform,
        )
        self._engine._add_opts(state, mode, _opts)


class QFontIconStore(QObject):
    # map of key -> (font_family, font_style)
    _LOADED_KEYS: ClassVar[dict[str, tuple[str, str]]] = {}

    # map of (font_family, font_style) -> character (char may include key)
    _CHARMAPS: ClassVar[dict[tuple[str, str | None], dict[str, str]]] = {}

    # singleton instance, use `instance()` to retrieve
    __instance: ClassVar[QFontIconStore | None] = None

    def __init__(self, parent: QObject | None = None) -> None:
        super().__init__(parent=parent)
        if tuple(cast(str, QT_VERSION).split(".")) < ("6", "0"):
            # QT6 drops this
            QApplication.setAttribute(Qt.ApplicationAttribute.AA_UseHighDpiPixmaps)

    @classmethod
    def instance(cls) -> QFontIconStore:
        if cls.__instance is None:
            cls.__instance = cls()
        return cls.__instance

    @classmethod
    def clear(cls) -> None:
        cls._LOADED_KEYS.clear()
        cls._CHARMAPS.clear()
        QFontDatabase.removeAllApplicationFonts()

    @classmethod
    def _key2family(cls, key: str) -> tuple[str, str]:
        """Return (family, style) given a font `key`."""
        key = key.split(".", maxsplit=1)[0]
        if key not in cls._LOADED_KEYS:
            from . import _plugins

            try:
                font_cls = _plugins.get_font_class(key)
                result = cls.addFont(
                    font_cls.__font_file__, key, charmap=dict(font_cls.__dict__)
                )
                if not result:  # pragma: no cover
                    raise Exception("Invalid font file")
                cls._LOADED_KEYS[key] = result
            except ValueError as e:
                raise ValueError(
                    f"Unrecognized font key: {key!r}.\n"
                    f"Known plugin keys include: {_plugins.available()}.\n"
                    f"Loaded keys include: {list(cls._LOADED_KEYS)}."
                ) from e
        return cls._LOADED_KEYS[key]

    @classmethod
    def _ensure_char(cls, char: str, family: str, style: str) -> str:
        """Make sure that `char` is a glyph provided by `family` and `style`."""
        if len(char) == 1 and ord(char) > 256:
            return char
        try:
            charmap = cls._CHARMAPS[(family, style)]
        except KeyError as e:
            raise KeyError(
                f"No charmap registered for font '{family} ({style})'"
            ) from e
        if char in charmap:
            # split in case the charmap includes the key
            return charmap[char].split(".", maxsplit=1)[-1]

        ident = _ensure_identifier(char)
        if ident in charmap:
            return charmap[ident].split(".", maxsplit=1)[-1]

        ident = f"{char!r} or {ident!r}" if char != ident else repr(ident)
        raise ValueError(f"Font '{family} ({style})' has no glyph with the key {ident}")

    @classmethod
    def key2glyph(cls, glyph_key: str) -> tuple[str, str, str | None]:
        """Return (char, family, style) given a `glyph_key`."""
        if "." not in glyph_key:
            raise ValueError("Glyph key must contain a period")
        font_key, char = glyph_key.split(".", maxsplit=1)
        family, style = cls._key2family(font_key)
        char = cls._ensure_char(char, family, style)
        return char, family, style

    @classmethod
    def addFont(
        cls, filepath: str, prefix: str, charmap: dict[str, str] | None = None
    ) -> tuple[str, str] | None:
        r"""Add font at `filepath` to the registry under `key`.

        If you'd like to later use a fontkey in the form of `key.some-name`, then
        `charmap` must be provided and provide a mapping for all of the glyph names
        to their unicode numbers. If a charmap is not provided, glyphs must be directly
        accessed with their unicode as something like `key.\\uffff`.

        Parameters
        ----------
        filepath : str
            Path to an OTF or TTF file containing the fonts
        prefix : str
            A key that will represent this font file when used for lookup.  For example,
            'fa5s' for 'Font-Awesome 5 Solid'.
        charmap : Dict[str, str], optional
            optional mapping for all of the glyph names to their unicode numbers.
            See note above.

        Returns
        -------
        Tuple[str, str], optional
            font-family and font-style for the file just registered, or None if
            something goes wrong.
        """
        if prefix in cls._LOADED_KEYS:
            warnings.warn(f"Prefix {prefix} already loaded", stacklevel=2)
            return None

        if not Path(filepath).exists():
            raise FileNotFoundError(f"Font file doesn't exist: {filepath}")
        if QApplication.instance() is None:
            raise RuntimeError("Please create QApplication before adding a Font")

        fontId = QFontDatabase.addApplicationFont(str(Path(filepath).absolute()))
        if fontId < 0:  # pragma: no cover
            warnings.warn(f"Cannot load font file: {filepath}", stacklevel=2)
            return None

        families = QFontDatabase.applicationFontFamilies(fontId)
        if not families:  # pragma: no cover
            warnings.warn(f"Font file is empty!: {filepath}", stacklevel=2)
            return None
        family: str = families[0]

        # in Qt6, everything becomes a static member
        QFd: QFontDatabase | type[QFontDatabase] = (
            QFontDatabase()
            if tuple(cast(str, QT_VERSION).split(".")) < ("6", "0")
            else QFontDatabase
        )

        styles = QFd.styles(family)
        style: str = styles[-1] if styles else ""
        if not QFd.isSmoothlyScalable(family, style):  # pragma: no cover
            warnings.warn(
                f"Registered font {family} ({style}) is not smoothly scalable. "
                "Icons may not look attractive.",
                stacklevel=2,
            )

        cls._LOADED_KEYS[prefix] = (family, style)
        if charmap:
            cls._CHARMAPS[(family, style)] = charmap
        return (family, style)

    def icon(
        self,
        glyph_key: str,
        *,
        scale_factor: float = DEFAULT_SCALING_FACTOR,
        color: ValidColor | None = None,
        opacity: float = 1,
        animation: Animation | None = None,
        transform: QTransform | None = None,
        states: dict[str, IconOptionDict | IconOpts] | None = None,
    ) -> QFontIcon:
        self.key2glyph(glyph_key)  # make sure it's a valid glyph_key
        default_opts = _IconOptions(
            glyph_key=glyph_key,
            scale_factor=scale_factor,
            color=color,
            opacity=opacity,
            animation=animation,
            transform=transform,
        )
        icon = QFontIcon(default_opts)
        for kw, options in (states or {}).items():
            if isinstance(options, IconOpts):
                options = default_opts._update(options).dict()
            icon.addState(*_norm_state_mode(kw), **options)
        return icon

    def setTextIcon(
        self, widget: QWidget, glyph_key: str, size: float | None = None
    ) -> None:
        """Sets text on a widget to a specific font & glyph.

        This is an alternative to setting a `QIcon` with a pixmap.  It may
        be easier to combine with dynamic stylesheets.
        """
        setText = getattr(widget, "setText", None)
        if not setText:  # pragma: no cover
            raise TypeError(f"Object does not a setText method: {widget}")

        glyph = self.key2glyph(glyph_key)[0]
        size = size or DEFAULT_SCALING_FACTOR
        size = size if size > 1 else widget.height() * size
        widget.setFont(self.font(glyph_key, int(size)))
        setText(glyph)

    def font(self, font_prefix: str, size: int | None = None) -> QFont:
        """Create QFont for `font_prefix`."""
        font_key, _ = font_prefix.split(".", maxsplit=1)
        family, style = self._key2family(font_key)
        font = QFont()
        font.setFamily(family)
        if style:
            font.setStyleName(style)
        if size:
            font.setPixelSize(int(size))
        return font


def _ensure_identifier(name: str) -> str:
    """Normalize string to valid identifier."""
    import keyword

    if not name:
        return ""

    # add _ to beginning of names starting with numbers
    if name[0].isdigit():
        name = f"_{name}"

    # add _ to end of reserved keywords
    if keyword.iskeyword(name):
        name += "_"

    # replace dashes and spaces with underscores
    name = name.replace("-", "_").replace(" ", "_")

    if not str.isidentifier(name):
        raise ValueError(f"Could not canonicalize name: {name!r}. (not an identifier)")
    return name



from __future__ import annotations

from inspect import signature
from typing import Callable


def get_max_args(func: Callable) -> int | None:
    """Return the maximum number of positional arguments that func can accept."""
    if not callable(func):
        raise TypeError(f"{func!r} is not callable")

    try:
        sig = signature(func)
    except Exception:
        return None

    max_args = 0
    for param in sig.parameters.values():
        if param.kind == param.VAR_POSITIONAL:
            return None
        if param.kind in {param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD}:
            max_args += 1
    return max_args



from qtpy.QtWidgets import QTextEdit

from superqt.utils import CodeSyntaxHighlight


def test_code_highlight(qtbot):
    widget = QTextEdit()
    qtbot.addWidget(widget)
    code_highlight = CodeSyntaxHighlight(widget, "python", "default")
    assert code_highlight.background_color == "#f8f8f8"
    widget.setText("from argparse import ArgumentParser")


def test_code_highlight_by_name(qtbot):
    widget = QTextEdit()
    qtbot.addWidget(widget)
    code_highlight = CodeSyntaxHighlight(widget, "Python Traceback", "monokai")
    assert code_highlight.background_color == "#272822"
    widget.setText("from argparse import ArgumentParser")



from unittest.mock import Mock

import pytest
from qtpy.QtCore import QObject, Signal

from superqt.utils import qdebounced, qthrottled
from superqt.utils._throttler import ThrottledCallable


def test_debounced(qtbot):
    mock1 = Mock()
    mock2 = Mock()

    @qdebounced(timeout=5)
    def f1() -> str:
        mock1()

    def f2() -> str:
        mock2()

    for _ in range(10):
        f1()
        f2()

    qtbot.wait(5)
    mock1.assert_called_once()
    assert mock2.call_count == 10


def test_debouncer_method(qtbot):
    class A(QObject):
        def __init__(self):
            super().__init__()
            self.count = 0

        def callback(self):
            self.count += 1

    a = A()
    assert all(not isinstance(x, ThrottledCallable) for x in a.children())
    b = qdebounced(a.callback, timeout=4)
    assert any(isinstance(x, ThrottledCallable) for x in a.children())
    for _ in range(10):
        b()

    qtbot.wait(5)

    assert a.count == 1


def test_debouncer_method_definition(qtbot):
    mock1 = Mock()
    mock2 = Mock()

    class A(QObject):
        def __init__(self):
            super().__init__()
            self.count = 0

        @qdebounced(timeout=4)
        def callback(self):
            self.count += 1

        @qdebounced(timeout=4)
        @staticmethod
        def call1():
            mock1()

        @staticmethod
        @qdebounced(timeout=4)
        def call2():
            mock2()

    a = A()
    assert all(not isinstance(x, ThrottledCallable) for x in a.children())
    for _ in range(10):
        a.callback(1)
        A.call1(34)
        a.call1(22)
        a.call2(22)
        A.call2(32)

    qtbot.wait(5)

    assert a.count == 1
    mock1.assert_called_once()
    mock2.assert_called_once()


def test_class_with_slots(qtbot):
    class A:
        __slots__ = ("count", "__weakref__")

        def __init__(self):
            self.count = 0

        @qdebounced(timeout=4)
        def callback(self):
            self.count += 1

    a = A()
    for _ in range(10):
        a.callback()

    qtbot.wait(5)
    assert a.count == 1


@pytest.mark.usefixtures("qapp")
def test_class_with_slots_except():
    class A:
        __slots__ = ("count",)

        def __init__(self):
            self.count = 0

        @qdebounced(timeout=4)
        def callback(self):
            self.count += 1

    with pytest.raises(TypeError, match="To use qthrottled or qdebounced"):
        A().callback()


def test_throttled(qtbot):
    mock1 = Mock()
    mock2 = Mock()

    @qthrottled(timeout=5)
    def f1() -> str:
        mock1()

    def f2() -> str:
        mock2()

    for _ in range(10):
        f1()
        f2()

    qtbot.wait(5)
    assert mock1.call_count == 2
    assert mock2.call_count == 10


@pytest.mark.parametrize("deco", [qthrottled, qdebounced])
def test_ensure_throttled_sig_inspection(deco, qtbot):
    mock = Mock()

    class Emitter(QObject):
        sig = Signal(int, int, int)

    @deco
    def func(a: int, b: int):
        """docstring"""
        mock(a, b)

    obj = Emitter()
    obj.sig.connect(func)

    # this is the crux of the test...
    # we emit 3 args, but the function only takes 2
    # this should normally work fine in Qt.
    # testing here that the decorator doesn't break it.
    with qtbot.waitSignal(func.triggered, timeout=1000):
        obj.sig.emit(1, 2, 3)
    mock.assert_called_once_with(1, 2)
    assert func.__doc__ == "docstring"
    assert func.__name__ == "func"



import math
from itertools import product
from typing import Any, Iterable
from unittest.mock import Mock

import pytest
from qtpy.QtCore import QEvent, QPoint, QPointF, Qt
from qtpy.QtWidgets import QStyle, QStyleOptionSlider

from superqt import QDoubleRangeSlider, QLabeledRangeSlider, QRangeSlider

from ._testutil import (
    _hover_event,
    _linspace,
    _mouse_event,
    _wheel_event,
    skip_on_linux_qt6,
)

ALL_SLIDER_COMBOS = list(
    product(
        [QDoubleRangeSlider, QRangeSlider, QLabeledRangeSlider],
        [Qt.Orientation.Horizontal, Qt.Orientation.Vertical],
    )
)
FLOAT_SLIDERS = [c for c in ALL_SLIDER_COMBOS if c[0] == QDoubleRangeSlider]


@pytest.mark.parametrize("cls, orientation", ALL_SLIDER_COMBOS)
def test_slider_init(qtbot, cls, orientation):
    slider = cls(orientation)
    assert slider.value() == (20, 80)
    assert slider.minimum() == 0
    assert slider.maximum() == 99
    slider.show()
    qtbot.addWidget(slider)


@pytest.mark.parametrize("cls, orientation", ALL_SLIDER_COMBOS)
def test_change_floatslider_range(cls, orientation, qtbot):
    sld = cls(orientation)
    qtbot.addWidget(sld)

    with qtbot.waitSignals([sld.rangeChanged, sld.valueChanged]):
        sld.setMinimum(30)

    assert sld.value()[0] == 30 == sld.minimum()
    assert sld.maximum() == 99

    with qtbot.waitSignal(sld.rangeChanged):
        sld.setMaximum(70)
    assert sld.value()[0] == 30 == sld.minimum()
    assert sld.value()[1] == 70 == sld.maximum()

    with qtbot.waitSignals([sld.rangeChanged, sld.valueChanged]):
        sld.setRange(40, 60)
    assert sld.value()[0] == 40 == sld.minimum()
    assert sld.maximum() == 60

    with qtbot.waitSignal(sld.valueChanged):
        sld.setValue([40, 50])
    assert sld.value()[0] == 40 == sld.minimum()
    assert sld.value()[1] == 50

    with qtbot.waitSignals([sld.rangeChanged, sld.valueChanged]):
        sld.setMaximum(45)
    assert sld.value()[0] == 40 == sld.minimum()
    assert sld.value()[1] == 45 == sld.maximum()


@pytest.mark.parametrize("cls, orientation", FLOAT_SLIDERS)
def test_float_values(cls, orientation, qtbot):
    sld = cls(orientation)
    qtbot.addWidget(sld)

    with qtbot.waitSignal(sld.rangeChanged):
        sld.setRange(0.1, 0.9)
    assert sld.minimum() == 0.1
    assert sld.maximum() == 0.9

    with qtbot.waitSignal(sld.valueChanged):
        sld.setValue([0.4, 0.6])
    assert sld.value() == (0.4, 0.6)

    with qtbot.waitSignal(sld.valueChanged):
        sld.setValue([0, 1.9])
    assert sld.value()[0] == 0.1 == sld.minimum()
    assert sld.value()[1] == 0.9 == sld.maximum()


@pytest.mark.parametrize("cls, orientation", ALL_SLIDER_COMBOS)
def test_position(cls, orientation, qtbot):
    sld = cls(orientation)
    qtbot.addWidget(sld)

    sld.setSliderPosition([10, 80])
    assert sld.sliderPosition() == (10, 80)


@pytest.mark.parametrize("cls, orientation", ALL_SLIDER_COMBOS)
def test_steps(cls, orientation, qtbot):
    sld = cls(orientation)
    qtbot.addWidget(sld)

    sld.setSingleStep(0.1)
    assert sld.singleStep() == 0.1

    sld.setSingleStep(1.5e20)
    assert sld.singleStep() == 1.5e20

    sld.setPageStep(0.2)
    assert sld.pageStep() == 0.2

    sld.setPageStep(1.5e30)
    assert sld.pageStep() == 1.5e30


@pytest.mark.parametrize("mag", list(range(4, 37, 4)) + list(range(-4, -37, -4)))
@pytest.mark.parametrize("cls, orientation", FLOAT_SLIDERS)
def test_slider_extremes(cls, orientation, qtbot, mag):
    sld = cls(orientation)
    qtbot.addWidget(sld)

    _mag = 10**mag
    with qtbot.waitSignal(sld.rangeChanged):
        sld.setRange(-_mag, _mag)
    for i in _linspace(-_mag, _mag, 10):
        sld.setValue((i, _mag))
        assert math.isclose(sld.value()[0], i, rel_tol=0.0001)
        sld.initStyleOption(QStyleOptionSlider())


@pytest.mark.parametrize("cls, orientation", ALL_SLIDER_COMBOS)
def test_ticks(cls, orientation, qtbot):
    sld = cls(orientation)
    qtbot.addWidget(sld)

    sld.setTickInterval(0.3)
    assert sld.tickInterval() == 0.3
    sld.setTickPosition(sld.TickPosition.TicksAbove)
    sld.show()


@pytest.mark.parametrize("cls, orientation", FLOAT_SLIDERS)
def test_press_move_release(cls, orientation, qtbot):
    sld = cls(orientation)
    qtbot.addWidget(sld)

    # this fail on vertical came with pyside6.2 ... need to debug
    # still works in practice, but test fails to catch signals
    if sld.orientation() == Qt.Orientation.Vertical:
        pytest.xfail()

    assert sld._pressedControl == QStyle.SubControl.SC_None

    opt = QStyleOptionSlider()
    sld.initStyleOption(opt)
    style = sld.style()
    hrect = style.subControlRect(
        QStyle.ComplexControl.CC_Slider, opt, QStyle.SubControl.SC_SliderHandle
    )
    handle_pos = sld.mapToGlobal(hrect.center())

    with qtbot.waitSignal(sld.sliderPressed):
        qtbot.mousePress(sld, Qt.MouseButton.LeftButton, pos=handle_pos)

    assert sld._pressedControl == QStyle.SubControl.SC_SliderHandle

    with qtbot.waitSignals([sld.sliderMoved, sld.valueChanged]):
        shift = (
            QPoint(0, -8)
            if sld.orientation() == Qt.Orientation.Vertical
            else QPoint(8, 0)
        )
        sld.mouseMoveEvent(_mouse_event(handle_pos + shift))

    with qtbot.waitSignal(sld.sliderReleased):
        qtbot.mouseRelease(sld, Qt.MouseButton.LeftButton, pos=handle_pos)

    assert sld._pressedControl == QStyle.SubControl.SC_None

    sld.show()
    with qtbot.waitSignal(sld.sliderPressed):
        qtbot.mousePress(sld, Qt.MouseButton.LeftButton, pos=handle_pos)


@skip_on_linux_qt6
@pytest.mark.parametrize("cls, orientation", FLOAT_SLIDERS)
def test_hover(cls, orientation, qtbot):
    sld = cls(orientation)
    qtbot.addWidget(sld)

    hrect = sld._handleRect(0)
    handle_pos = QPointF(sld.mapToGlobal(hrect.center()))

    assert sld._hoverControl == QStyle.SubControl.SC_None

    sld.event(_hover_event(QEvent.Type.HoverEnter, handle_pos, QPointF(), sld))
    assert sld._hoverControl == QStyle.SubControl.SC_SliderHandle

    sld.event(
        _hover_event(QEvent.Type.HoverLeave, QPointF(-1000, -1000), handle_pos, sld)
    )
    assert sld._hoverControl == QStyle.SubControl.SC_None


@pytest.mark.parametrize("cls, orientation", FLOAT_SLIDERS)
def test_wheel(cls, orientation, qtbot):
    sld = cls(orientation)
    qtbot.addWidget(sld)

    with qtbot.waitSignal(sld.valueChanged):
        sld.wheelEvent(_wheel_event(120))

    sld.wheelEvent(_wheel_event(0))


def _assert_types(args: Iterable[Any], type_: type):
    # sourcery skip: comprehension-to-generator
    assert all(isinstance(v, type_) for v in args), "invalid type"


@pytest.mark.parametrize("cls, orientation", ALL_SLIDER_COMBOS)
def test_rangeslider_signals(cls, orientation, qtbot):
    sld = cls(orientation)
    qtbot.addWidget(sld)

    type_ = float if cls == QDoubleRangeSlider else int

    mock = Mock()
    sld.valueChanged.connect(mock)
    with qtbot.waitSignal(sld.valueChanged):
        sld.setValue((20, 40))
    mock.assert_called_once_with((20, 40))
    _assert_types(mock.call_args.args, tuple)
    _assert_types(mock.call_args.args[0], type_)

    mock = Mock()
    sld.rangeChanged.connect(mock)
    with qtbot.waitSignal(sld.rangeChanged):
        sld.setMinimum(3)
    mock.assert_called_once_with(3, 99)
    _assert_types(mock.call_args.args, type_)

    mock.reset_mock()
    with qtbot.waitSignal(sld.rangeChanged):
        sld.setMaximum(15)
    mock.assert_called_once_with(3, 15)
    _assert_types(mock.call_args.args, type_)

    mock.reset_mock()
    with qtbot.waitSignal(sld.rangeChanged):
        sld.setRange(1, 2)
    mock.assert_called_once_with(1, 2)
    _assert_types(mock.call_args.args, type_)



from abc import ABC, abstractmethod
from typing import Optional

from qtpy.QtCore import QRectF, QTimer
from qtpy.QtGui import QPainter
from qtpy.QtWidgets import QWidget


class Animation(ABC):
    """Base icon animation class."""

    def __init__(self, parent_widget: QWidget, interval: int = 10, step: int = 1):
        self.parent_widget = parent_widget
        self.timer = QTimer()
        self.timer.timeout.connect(self._update)  # type: ignore
        self.timer.setInterval(interval)
        self._angle = 0
        self._step = step

    def _update(self):
        if self.timer.isActive():
            self._angle += self._step
            self.parent_widget.update()

    @abstractmethod
    def animate(self, painter: QPainter):
        """Setup and start the timer for the animation."""


class spin(Animation):
    """Animation that smoothly spins an icon."""

    def animate(self, painter: QPainter):
        if not self.timer.isActive():
            self.timer.start()

        mid = QRectF(painter.viewport()).center()
        painter.translate(mid)
        painter.rotate(self._angle % 360)
        painter.translate(-mid)


class pulse(spin):
    """Animation that spins an icon in slower, discrete steps."""

    def __init__(self, parent_widget: Optional[QWidget] = None):
        super().__init__(parent_widget, interval=200, step=45)



"""A test module for testing collapsible"""

from qtpy.QtCore import QEasingCurve, Qt
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QPushButton, QStyle, QWidget

from superqt import QCollapsible


def _get_builtin_icon(name: str) -> QIcon:
    """Get a built-in icon from the Qt library."""
    widget = QWidget()
    try:
        pixmap = getattr(QStyle.StandardPixmap, f"SP_{name}")
    except AttributeError:
        pixmap = getattr(QStyle, f"SP_{name}")

    return widget.style().standardIcon(pixmap)


def test_checked_initialization(qtbot):
    """Test simple collapsible"""
    wdg1 = QCollapsible("Advanced analysis")
    wdg1.expand(False)
    assert wdg1.isExpanded()
    assert wdg1._content.maximumHeight() > 0

    wdg2 = QCollapsible("Advanced analysis")
    wdg1.collapse(False)
    assert not wdg2.isExpanded()
    assert wdg2._content.maximumHeight() == 0


def test_content_hide_show(qtbot):
    """Test collapsible with content"""

    # Create child component
    collapsible = QCollapsible("Advanced analysis")
    for i in range(10):
        collapsible.addWidget(QPushButton(f"Content button {i + 1}"))

    collapsible.collapse(False)
    assert not collapsible.isExpanded()
    assert collapsible._content.maximumHeight() == 0

    collapsible.expand(False)
    assert collapsible.isExpanded()
    assert collapsible._content.maximumHeight() > 0


def test_locking(qtbot):
    """Test locking collapsible"""
    wdg1 = QCollapsible()
    assert wdg1.locked() is False
    wdg1.setLocked(True)
    assert wdg1.locked() is True
    assert not wdg1.isExpanded()

    wdg1._toggle_btn.setChecked(True)
    assert not wdg1.isExpanded()

    wdg1._toggle()
    assert not wdg1.isExpanded()

    wdg1.expand()
    assert not wdg1.isExpanded()

    wdg1._toggle_btn.setChecked(False)
    assert not wdg1.isExpanded()

    wdg1.setLocked(False)
    wdg1.expand()
    assert wdg1.isExpanded()
    assert wdg1._toggle_btn.isChecked()


def test_changing_animation_settings(qtbot):
    """Quick test for changing animation settings"""
    wdg = QCollapsible()
    wdg.setDuration(600)
    wdg.setEasingCurve(QEasingCurve.Type.InElastic)
    assert wdg._animation.easingCurve() == QEasingCurve.Type.InElastic
    assert wdg._animation.duration() == 600


def test_changing_content(qtbot):
    """Test changing the content"""
    content = QPushButton()
    wdg = QCollapsible()
    wdg.setContent(content)
    assert wdg._content == content


def test_changing_text(qtbot):
    """Test changing the content"""
    wdg = QCollapsible()
    wdg.setText("Hi new text")
    assert wdg.text() == "Hi new text"
    assert wdg._toggle_btn.text() == "Hi new text"


def test_toggle_signal(qtbot):
    """Test that signal is emitted when widget expanded/collapsed."""
    wdg = QCollapsible()
    with qtbot.waitSignal(wdg.toggled, timeout=500):
        qtbot.mouseClick(wdg._toggle_btn, Qt.LeftButton)

    with qtbot.waitSignal(wdg.toggled, timeout=500):
        wdg.expand()

    with qtbot.waitSignal(wdg.toggled, timeout=500):
        wdg.collapse()


def test_getting_icon(qtbot):
    """Test setting string as toggle button."""
    wdg = QCollapsible("test")
    assert isinstance(wdg.expandedIcon(), QIcon)
    assert isinstance(wdg.collapsedIcon(), QIcon)


def test_setting_icon(qtbot):
    """Test setting icon for toggle button."""
    icon1 = _get_builtin_icon("ArrowRight")
    icon2 = _get_builtin_icon("ArrowDown")
    wdg = QCollapsible("test", expandedIcon=icon1, collapsedIcon=icon2)
    assert wdg._expanded_icon == icon1
    assert wdg._collapsed_icon == icon2


def test_setting_symbol_icon(qtbot):
    """Test setting string as toggle button."""
    wdg = QCollapsible("test")
    icon1 = wdg._convert_string_to_icon("+")
    icon2 = wdg._convert_string_to_icon("-")
    wdg.setCollapsedIcon(icon=icon1)
    assert wdg._collapsed_icon == icon1
    wdg.setExpandedIcon(icon=icon2)
    assert wdg._expanded_icon == icon2



__all__ = (
    "CodeSyntaxHighlight",
    "create_worker",
    "ensure_main_thread",
    "ensure_object_thread",
    "FunctionWorker",
    "GeneratorWorker",
    "new_worker_qthread",
    "qdebounced",
    "QMessageHandler",
    "QSignalDebouncer",
    "QSignalThrottler",
    "qthrottled",
    "signals_blocked",
    "thread_worker",
    "WorkerBase",
    "exceptions_as_dialog",
)

from ._code_syntax_highlight import CodeSyntaxHighlight
from ._ensure_thread import ensure_main_thread, ensure_object_thread
from ._errormsg_context import exceptions_as_dialog
from ._message_handler import QMessageHandler
from ._misc import signals_blocked
from ._qthreading import (
    FunctionWorker,
    GeneratorWorker,
    WorkerBase,
    create_worker,
    new_worker_qthread,
    thread_worker,
)
from ._throttler import QSignalDebouncer, QSignalThrottler, qdebounced, qthrottled
